
/* Initng, a next generatind sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#undef FILENAME
#define FILENAME "active_db.c"
#include "initng.h"

#include <sys/types.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <alloca.h>
#include "active_db.h"
#include "service_db.h"
#include "process_db.h"

/* Tis is the path to the big global active_db */

/*active_h *active_database[MAX_ACTIVE];*/
active_h **active_database = NULL;

void actictive_db_change_active_h(active_h * from, active_h * to);
void actictive_db_change_service_h(service_h * from, service_h * to);

const char *active_status_string[] = {
    "STATUS_NONE",
    "STATUS_LOADING",
    "STATUS_STARTING",
    "STATUS_WAITING_FOR_START_DEP",
    "STATUS_START_DEP_MET",
    "STATUS_FAIL_STARTING",
    "STATUS_RUNNING",
    "STATUS_DONE",
    "STATUS_RESPAWNING",
    "STATUS_STOPING",
    "STATUS_WAITING_FOR_STOP_DEP",
    "STATUS_STOP_DEP_MET",
    "STATUS_FAIL_STOPING",
    "STATUS_STOPED",
    "STATUS_STOP_MARKED",
    NULL
};

int next_active_to_alloc = 0;

/* return index of service in active data structure or -1 if not found */
active_h *active_db_find_by_name(char *service)
{
    int i;

    for (i = 0; i < next_active_to_alloc; i++)
    {
        if (active_database[i] && active_database[i]->name
            && strcasecmp(active_database[i]->name, service) == 0)
            return (active_database[i]);
    }

    return NULL;
}

/* return index of service in active data structure or -1 if not found */
active_h *active_db_find_by_service_h(service_h * service)
{
    int i;

    for (i = 0; i < next_active_to_alloc; i++)
    {
        if (active_database[i] && active_database[i]->from_service
            && active_database[i]->from_service == service)
            return (active_database[i]);
    }

    return NULL;
}

/*int active_db_reget_data(active_h *service) {
    if(!(service->from_service=get_service(service->name))) {
	printf(" ** FAIL active_db_reget_data(%s): coud not load service.\n",service->name);
	return(FAIL);
    }
    return(TRUE);	
}*/

/* add active to data structure, send path to a active_h, return index or -1 */
int active_db_add(active_h ** add_this)
{
    active_h *tmp_active = 0;
    int i = 0;

    if (!*add_this)
        return (FALSE);

    /* Look for duplicate - both name and type */
    if ((tmp_active = active_db_find_by_name((*add_this)->name)))
    {
#ifdef DEBUG
        printf(" ** " FILENAME " active_db_add(): duplicate here\n");
#endif
        active_db_free(*add_this);
        *add_this = NULL;
        return (FALSE);
    }

    /* if it is the first one ever, it may not be null */
    if (!next_active_to_alloc)
        active_database[0] = NULL;

    /* cout up all pointer looking for a free one  and less then maximum */
    for (i = 0; (i <= next_active_to_alloc && i <= MAX_ACTIVE); i++)
    {
        /* if this post is free */
        if (!active_database[i])
        {
            /* point this post to the allocated area got */
            active_database[i] = *add_this;

            /* if this is the last post */
            if (i == next_active_to_alloc)
            {
                /* count up next active top alloc */
                next_active_to_alloc++;

                /* set next one to zero */
                active_database[next_active_to_alloc] = NULL;

            }
#ifdef DEBUG
            printf(" ** " FILENAME
                   " active_db_add(%s): added entry in active_database[%i].\n",
                   active_database[i]->from_service->name, i);
#endif
            return (TRUE);
        }
    }

    printf(" ** FAIL " FILENAME
           " active_db_add(%s): unknown error, maby max actives reaced.\n",
           active_database[i]->from_service->name);
    active_db_free(*add_this);
    *add_this = NULL;
    return (FALSE);
}

/* free som values in this one */
void active_db_free(active_h * pf)
{
    /* dumb check */
    if (!pf)
        return;

#ifdef DEBUG
    printf("FREEING %s.\n", (*pf)->name);
#endif
    /* reset to null i complete database */
    actictive_db_change_active_h(pf, NULL);

    /* try free name */
    if (pf->name)
        free(pf->name);

    /* try free process structs */
    process_db_free(&pf->start_process);
    process_db_free(&pf->stop_process);

    /* free dependncy registry */
    if (pf->depends)
        free(pf->depends);

    /* free active_h */
    free(pf);

}

/* clear database */
void active_db_free_all(void)
{
    int i;

    for (i = 0; i < next_active_to_alloc; i++)
    {
        if (active_database)
            if (active_database[i])
            {
                active_db_free(active_database[i]);
                /* active_database[i]=NULL; */
            }
    }
    if (active_database)
        free(active_database);
    active_database = NULL;
}

/* compensate time */
void active_db_compensate_time(time_t screw)
{
    int i;

    for (i = 0; i < next_active_to_alloc; i++)
    {
        if (active_database && active_database[i]
            && active_database[i]->time_got_status)
        {
            active_database[i]->time_got_status =
                active_database[i]->time_got_status + screw;
        }
    }
}

/* update all active_h pointers in hole active_db */
void actictive_db_change_active_h(active_h * from, active_h * to)
{
    int i = 0, j = 0;

    if (!from)
        return;

    for (i = 0; i < next_active_to_alloc; i++)
    {
        if (active_database[i])
        {
            if (active_database[i]->depends)
                for (j = 0; active_database[i]->depends[j]; j++)
                    if (active_database[i]->depends[j] == from)
                        active_database[i]->depends[j] = to;

/*	    if(active_database[i]->r_depends)
		for(j=0;active_database[i]->r_depends[j];j++)
		    if(active_database[i]->r_depends[j]==from)
			active_database[i]->r_depends[j]=to;*/
            if (active_database[i] == from)
                active_database[i] = to;
        }
    }
}

/* update all service_h pointers in hole active_db */
void active_db_change_service_h(service_h * from, service_h * to)
{
    int i = 0;

    if (!from)
        return;
    /* there migt not be a database */
    if (!active_database)
        return;

    for (i = 0; i < next_active_to_alloc; i++)
        if (active_database[i] && active_database[i]->from_service == from)
            active_database[i]->from_service = to;
}

/* clear database */
void active_db_init(void)
{
    /*int i; */
    active_database = calloc(MAX_ACTIVE + 1, sizeof(active_h));
#ifdef ZERO_MEMORY
    /* calloc zeros anyway */
    for (i = 0; i < MAX_ACTIVE; i++)
    {
        active_database[i] = NULL;
    }
#endif
}

/* return next active entry */
active_h *active_db_next(active_h * current)
{
    int i = 0;

    /* if no current is found, return first active found */
    if (!current)
        for (i = 0; i < next_active_to_alloc; i++)
        {
            if (active_database[i])
            {
#ifdef DEBUG
                printf(" ** " FILENAME
                       " active_db_next(): Returning first active pointer... \n");
#endif
                return (active_database[i]);
            }
        }

    /* if previous did not found entry */
    if (i >= next_active_to_alloc)
    {
#ifdef DEBUG
        printf(" ** " FILENAME
               " active_db_next(): Did not find first free pointer... \n");
#endif
        return (NULL);
    }

    /* step forward to the one we got */
    for (i = 0; i < next_active_to_alloc; i++)
    {
        if (active_database[i] == current)
            break;
    }

    /* this funtions returns next pointer */
    i++;

    /* if previous did not found entry */
    if (i >= next_active_to_alloc)
    {
#ifdef DEBUG
        printf(" ** " FILENAME " active_db_next(): This was last... \n");
#endif
        return (NULL);
    }

    /* step forward, to we find next entry */
    for (; i < next_active_to_alloc; i++)
        if (active_database[i])
            return (active_database[i]);

    /* fail */
#ifdef DEBUG
    printf(" ** FAIL " FILENAME
           " active_db_next(): check this out, end of func! \n");
#endif
    return (NULL);
}

/* return next active entry */
active_h *active_db_next_dep(active_h * depend, active_h * current)
{
    int i = 0;
    int j = 0;

    /* if no current is found, return first active found */
    if (!current)
        for (i = 0; i < next_active_to_alloc; i++)
        {
            if (active_database[i] && active_database[i]->depends
                && active_database[i] != depend)
            {
                for (j = 0; active_database[i]->depends[j]; j++)
                {
                    if (active_database[i]->depends[j] == depend)
                        return (active_database[i]);
                }
            }
        }
    /* if previous did not found entry */
    if (i >= next_active_to_alloc)
    {
#ifdef DEBUG
        printf(" ** " FILENAME
               " active_db_next_dep(): Did not find first free pointer... \n");
#endif
        return (NULL);
    }

    /* step forward to the one we got */
    for (i = 0; i < next_active_to_alloc; i++)
    {
        if (active_database[i] == current)
            break;
    }

    /* this funtions returns next pointer */
    i++;

    /* if previous did not found entry */
    if (i >= next_active_to_alloc)
    {
#ifdef DEBUG
        printf(" ** " FILENAME " active_db_next_dep(): This was last... \n");
#endif
        return (NULL);
    }

    /* step forward, to we find next entry */
    for (; i < next_active_to_alloc; i++)
        if (active_database[i] && active_database[i]->depends
            && active_database[i] != depend)
            for (j = 0; active_database[i]->depends[j]; j++)
                if (active_database[i]->depends[j] == depend)
                    return (active_database[i]);

    /* fail */
#ifdef DEBUG
    printf(" ** FAIL " FILENAME
           " active_db_next_dep(): check this out, end of func! \n");
#endif
    return (NULL);
}

/* active_db_count counts a type, if null, cout all */
int active_db_count(e_a_status a_status_to_count)
{
    int i = 0;                  /* for counter */
    int counter = 0;            /* actives counter */

    /* shall we count a specific? */
    if (a_status_to_count)
    {
        /* count specific */
        for (i = 0; i < next_active_to_alloc; i++)
        {
            if (active_database[i]
                && active_database[i]->a_status == a_status_to_count)
                counter++;
        }
    }
    else
    {
        /* count all */
        for (i = 0; i < next_active_to_alloc; i++)
        {
            if (active_database[i] &&
                active_database[i]->a_status != FAIL_STARTING &&
                active_database[i]->a_status != FAIL_STOPING)
                counter++;
        }
    }

    /* return counter */
    return (counter);
}

/* returns pointer to active_h process belongs to, and sets process_type */
active_h *active_db_find_by_pid(pid_t pid)
{
    int i = 0;

    for (i = 0; i < next_active_to_alloc; i++)
    {
        if (active_database[i])
        {
            if (active_database[i]->start_process)
                if (active_database[i]->start_process->pid == pid)
                {
                    return (active_database[i]);
                }
            if (active_database[i]->stop_process)
                if (active_database[i]->stop_process->pid == pid)
                {
                    return (active_database[i]);
                }
        }
    }
    return (NULL);
}

/* creats a new active_h att new_active */
int active_db_new(active_h ** new_active, char *name)
{
    /* allocate a new active entry */
    *new_active = (active_h *) calloc(1, sizeof(active_h));     /* Allocate memory for a new active */

    /* out of memory? */
    if (!*new_active)
    {
        printf(" ** FAIL " FILENAME
               " active_db_new(): Unable to allocate active, out of memory?\n");
        return (FALSE);
    }

    (*new_active)->name = strdup(name);

    return (TRUE);

}

/* reset a active db */
void active_db_reset(active_h * new_active)
{
    new_active->time_got_status = 0;
    new_active->a_status = NONE;

    if (new_active->start_process)
        process_db_free(&new_active->start_process);
    if (new_active->stop_process)
        process_db_free(&new_active->stop_process);
}

const char *active_db_get_status_string(e_a_status status)
{
    if (status && status <= 9)
        return (active_status_string[status]);
    else
        return (active_status_string[0]);
}

/* calculate percent of processes started */
int active_db_percent_started(void)
{
    int i = 0;
    int starting = 0;
    int up = 0;
    float tmp = 0;

    for (i = 0; i < next_active_to_alloc; i++)
        if (active_database[i])
            switch (active_database[i]->a_status)
            {
                case STARTING:
                case WAITING_FOR_START_DEP:
                case START_DEP_MET:
                case LOADING:
#ifdef DEBUG
                    printf("%s: starting\n", active_database[i]->name);
#endif
                    starting++;
                    continue;
                case DONE:
                case RUNNING:
#ifdef DEBUG
                    printf("%s: up\n", active_database[i]->name);
#endif
                    up++;
                    starting++;
                    continue;
                default:
                    continue;
            }

#ifdef DEBUG
    printf("up: %i   starting: %i\n", up, starting);
#endif

    /* if no one starting */
    if (starting <= 0)
        return (100);

    if (up > 0)
    {
        tmp = 100 * (float) up / (float) starting;
#ifdef DEBUG
        printf("up/starting: %f percent: %i\n\n",
               (float) up / (float) starting, (int) tmp);
#endif
        return ((int) tmp);
    }
    return 0;
}

char *active_db_get_string(e_sdt type, active_h * from_active)
{
    return (service_db_get_string(type, from_active->from_service));
}

char *active_db_get_next_string(e_sdt type, active_h * from_active,
                                char *last)
{
    return (service_db_get_next_string
            (type, from_active->from_service, last));
}

void active_db_set_string(e_sdt type, active_h * from_active, char *string)
{
    service_db_set_string(type, from_active->from_service, string);
}

int active_db_is(e_sdt type, active_h * from_active)
{
    return (service_db_is(type, from_active->from_service));
}

int active_db_count_type(e_sdt type, active_h * from_active)
{
    return (service_db_count_type(type, from_active->from_service));
}

void active_db_set(e_sdt type, active_h * from_active)
{
    service_db_set(type, from_active->from_service);
}

void active_db_remove(e_sdt type, active_h * from_active)
{
    service_db_remove(type, from_active->from_service);
}
