
/* Initng, a next generatind sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#undef FILENAME
#define FILENAME "main.c"
#include "initng.h"

/* #define DEBUG */
#include <time.h>               /* time() */
#include <fcntl.h>              /* fcntl() */
#include <sys/un.h>             /* memmove() strcmp() */
#include <sys/wait.h>           /* waitpid() sa */
#include <linux/kd.h>           /* KDSIGACCEPT */
#include <sys/ioctl.h>          /* ioctl() */
#include <stdlib.h>             /* malloc() free() exit() */
#include <sys/reboot.h>         /* reboot() RB_DISABLE_CAD */
#include <termios.h>

#include "colorprint.h"
#include "main.h"
#include "initng_signal.h"
#include "initng_history.h"
#include "remote_control.h"
#include "initng_handler.h"
#include "initng_execute.h"     /* new_environ() */
#include "active_db.h"
char **Argv;
int Argc;

#undef printf
extern int printf(const char *format, ...);

int i_am_init = FALSE;
int interupt = FALSE;
h_then when_out = THEN_QUIT;
char *runlevel = NULL;

static void ready_to_quit(void);
static void su_login(void);
static void start_initiating_service(void);

/* Poll if there is any processes left, and if not quit initng */
static void ready_to_quit(void)
{
#ifdef REALY_DEBUG
    /* debug counting of remaining processes */
    printf("\n ** " FILENAME " ready_to_quit(), %i process(ess) remaining.\n",
           active_db_count(0));
    {
        active_h *tmp = NULL;

        while ((tmp = active_db_next(tmp)))
        {
            printf(" ********  %s status: %i left    *******\n", tmp->name,
                   tmp->a_status);
        }
    }
#endif

    /* if last process ded, quit initng */
    if ((active_db_count(0)) < 1)
    {
        /* print on screen */
        out_all_services_exited();
        if (runlevel)
            printf("\n\nRestarting from service/runlevel: %s\n\n", runlevel);
        switch (when_out)
        {
            default:
            case THEN_QUIT:
                exit_initng(0);
                break;
            case THEN_RESTART:
                restart_initng();
                break;
            case THEN_START_DEFAULT:
                /* start first service */
                start_initiating_service();
                break;
            case THEN_SULOGIN:
                su_login();
                break;
            case THEN_NEW_RUNLEVEL:
                if (runlevel)
                {
                    /* if runlevel is set to sulogin, do that */
                    if ((strcasecmp(runlevel, "sulogin") == 0))
                        su_login();
                    else
                        start_new_service_named(runlevel);
                }
                break;
        }
    }
}

/* exiting on an initialization failure is not a good idea for init */
static void su_login(void)
{
    char *argv[] = { "sulogin", 0 };
    execve("/sbin/sulogin", argv, new_environ(NULL));   /* launch sulogin */
    _exit(1);
}

/* find initziating services */
static void start_initiating_service(void)
{
    int i;
    int a_count = 0;            /* counts orders from argv to start */

    /* sheck with argv witch service to start initziating */
    for (i = 1; i < Argc; i++)
    {
        /* if suceed to load this service */
        if (start_new_service_named(Argv[i]))
            a_count++;
    }
}

/*    %%%%%%%%%%%%%%%%%%%%%        main()      %%%%%%%%%%%%%%%%%%%   */
#ifdef BUSYBOX
int init_main(int argc, char *argv[])
{
#else
int main(int argc, char *argv[])
{
#endif
    time_t last;                /* save the time here */
    int got_remote = FALSE;

#ifdef DEBUG
    int loop_counter = 0;       /* a common used for counter */

    printf(" ** " FILENAME " main();\n");
#endif

    /* set nice level */
    /* nice(INITNG_NICELEVEL); */

    /* Tell the user he is welcome */
    out_welcome();

    /* get the time */
    last = time(0);

    /* Null */
    service_db_init();
    history_db_init();
    active_db_init();

    /* Pointer to starting arguments, this makes them GLOBAL */
    Argv = argv;
    Argc = argc;

    /* Look if we are the first process born in the system */
    if (getpid() == 1)
    {
        int fd;                 /* /dev/console */

        i_am_init = TRUE;
        when_out = THEN_RESTART;
        reboot(RB_DISABLE_CAD); /* Disable Ctrl + Alt + Delete */

        if ((fd = open("/dev/console", O_RDWR | O_NOCTTY)))
        {                       /*O_NOCTTY=dont control tty */
            ioctl(fd, KDSIGACCEPT, SIGWINCH);   /* Accept kdb generated signals     */
            close(fd);          /* Like Ctrl + Alt + Delete signal? */
        }
        else
        {
            ioctl(0, KDSIGACCEPT, SIGWINCH);    /* Accept kdb generated signals */
        }

        /* this code is unknwon for me, i hope it does better, than worse */
        {
            struct termios tty;
            int fd;

            if ((fd = open("/dev/console", O_RDWR | O_NOCTTY)) < 0)
            {
                printf("can't open console\n");
            }

            (void) tcgetattr(fd, &tty);

            tty.c_cflag &= CBAUD | CBAUDEX | CSIZE | CSTOPB | PARENB | PARODD;
            tty.c_cflag |= HUPCL | CLOCAL | CREAD;

            tty.c_cc[VINTR] = 3;        /* ctrl('c') */
            tty.c_cc[VQUIT] = 28;       /* ctrl('\\') */
            tty.c_cc[VERASE] = 127;
            tty.c_cc[VKILL] = 24;       /* ctrl('x') */
            tty.c_cc[VEOF] = 4; /* ctrl('d') */
            tty.c_cc[VTIME] = 0;
            tty.c_cc[VMIN] = 1;
            tty.c_cc[VSTART] = 17;      /* ctrl('q') */
            tty.c_cc[VSTOP] = 19;       /* ctrl('s') */
            tty.c_cc[VSUSP] = 26;       /* ctrl('z') */

            /*
             *      Set pre and post processing
             */
            tty.c_iflag = IGNPAR | ICRNL | IXON | IXANY;
            tty.c_oflag = OPOST | ONLCR;
            tty.c_lflag = ISIG | ICANON | ECHO | ECHOCTL | ECHOPRT | ECHOKE;

            /*
             *      Now set the terminal line.
             *      We don't care about non-transmitted output data
             *      and non-read input data.
             */
            (void) tcsetattr(fd, TCSANOW, &tty);
            (void) tcflush(fd, TCIOFLUSH);
            (void) close(fd);
        }

    }

    /* enable signal requests */
    enable_signals();

    /* enable control tru /in and /out */
    got_remote = remote_control_open(i_am_init);

    /* if no starting services found, use default */
    if (!runlevel)
    {
        if (i_am_init)
        {
            runlevel = strdup("default");
        }
        else
        {
            runlevel = strdup("fake-default");
        }
    }

    start_new_service_named(runlevel);

    /* start first service */
    start_initiating_service();

    /*       %%%%%%%%%%%%%%%%%%%%%%%%%%%%%% MAIN LOOP %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%         */
    for (;;)
    {
        time_t now;             /* save the starting time here */

#ifdef DEBUG
        /* DEBUG print the loop count */
        printf("\n\n ** " FILENAME " main():mainloop: %i\n", loop_counter++);
#endif
        /* look at the watch */
        now = time(0);

        /* Check som global variables, set by sigactions */
        /* If we got an interupt load ctrlaltdel */
        if (initng_signal_got_ctrl_alt_del)
        {
            initng_signal_got_ctrl_alt_del = 0;
            start_new_service_named("ctrlaltdel");
        }
        /* If we got a keybord request load kbreq */
        if (initng_signal_got_keyboard_request)
        {
            initng_signal_got_keyboard_request = 0;
            start_new_service_named("kbreq");
        }

        /* IF now is befor last */
        if (now < last)
        {
#ifdef DEBUG
            printf(" ** " FILENAME
                   " main():mainloop: The clock was reset. Compensating.....\n");
#endif
            /* call function to compensate whole process_db */
            active_db_compensate_time(last - now);
            history_db_compensate_time(last - now);
        }
        /* if we have moved an hour ahed, we have to fix this, one hour at the time */
        while (now - last >= 3600)
        {
            active_db_compensate_time(3600);
            history_db_compensate_time(3600);
            last = last + 3600;
        }
        /* remember last time */
        last = now;

        /* Try start processes that waits for deps, the dep might be true now */

        /* try start all deps waiting, until db stabilisating */
        /* miniloop */
        while (interupt)
        {
            interupt = FALSE;
#ifdef DEBUG
            printf(" ** " FILENAME
                   " main():  *miniloop* ( will continue as log as interupt is set )\n");
#endif
            /* run try start_wait_deps as many times it want */
            try_start_wait_deps();
#ifdef DEBUG
            printf("try_start_wait_deps(): done!\n");
#endif
        }

        /* lok if it is any process that wont DIE */
        /*term_stoping_services(); */

        /* try to open remote control connection if faild before */
        if (!got_remote)
            got_remote = remote_control_open(i_am_init);

        /* check if there are any running processes left or quit */
        ready_to_quit();

        /* Look if anything is happening on the control chanel
         *  - wait for 60seconds */
        if (!interupt)
            remote_control_poll(60000);
    }
}

void restart_initng(void)
{
    printf("\n\n\n          R E S T A R T I N G,  (Realy hot reboot)\n\n");
    execve("/sbin/initng", Argv, environ);
}

/* end of life */
void exit_initng(int i)
{
#ifdef REALY_DEBUG
    printf(" ** " FILENAME " exit_initng();\n");
#endif
    /* flush history log db */
    history_free_all();
    /* close control fifos */
    remote_control_close();
    /* clean process_db */
    active_db_free_all();
    /* clean service db */
    service_db_free_all();
    if(runlevel) free(runlevel);
    /* exit program */
    if (i != 99)
        exit(i);
}
